package Server;
import java.net.InetSocketAddress;
import java.nio.ByteBuffer;
import java.nio.channels.ServerSocketChannel;
import java.nio.channels.SocketChannel;
import java.util.ArrayList;
import simsa.xml.XMLParser;

/** Administration console for the server.  This class
 *  contains the initial reference to the server.  It
 *  provides the functionality for loading, and unloading
 *  modules and viewing server statistics.
 *  2006-08-15 Kyle Reed
 */
public class AdminConsole extends Thread {
  private Server server;
  private Thread remoteConsole;
  ServerSocketChannel remoteAdminServer;
  private SocketChannel remoteAdminClient;
  private boolean terminate;
  ByteBuffer bb;
  byte[] ba;
  
  private XMLParser oxp;
  
  ArrayList<String> vo = new ArrayList<String>();
  
  /** Initialize the parser and start the server.
   *  2006-08-15 Kyle Reed
   */
  public AdminConsole() {
    server = new Server();
    vo.add("loadmod");    //0
    vo.add("killmod");    //1
    vo.add("killallmod"); //2
    vo.add("showmod");    //3
    vo.add("showclient"); //4
    vo.add("killclient"); //5
    vo.add("quit");       //6
    vo.add("help");       //7
    vo.add("killserver"); //8
    oxp = new XMLParser(vo);
  }
  /** The main AdminConsole loop that processes input from the
   *  admin terminal.  Loads and unloads modules and clients and
   *  gives server statistics. 2006-08-15 Kyle Reed
   */
  public void run() {
    try {
      remoteAdminServer = ServerSocketChannel.open();
      remoteAdminServer.socket().bind(new InetSocketAddress(12344));
      String inputString = "";
      terminate = false;
      
      while(server.isAlive() && remoteAdminServer.isOpen() && !terminate) {
        remoteAdminClient = remoteAdminServer.accept();
        System.out.println("Admin client: " +
            remoteAdminClient.socket().getLocalAddress().toString()
            + " connected: " + new java.util.Date().toString());
        writeRemote("SIMSA Server Administration Console");
        writeRemote("Type \'<help></help>\' for a list of commands. Type \'<quit></quit>\' to quit.");
        
        do {
          bb = ByteBuffer.allocate(server.MAX_BUFFER_SIZE);
          bb.clear();
          int readNum;
          readNum = remoteAdminClient.read(bb);
          bb.flip();
          ba = bb.array();
          
          //The read function returns -1 on EOF (Stream closed)
          if(readNum == -1) throw new Exception("Connection Closed.");
          
          for(int i = 0; i < readNum; i++) {
            oxp.parse((char)ba[i]);
            if(oxp.hasData()) {
              String data = oxp.getData();
              switch(oxp.getTagType()) {
                case 0: //load mod
                  writeRemoteRaw(server.loadModule(data));
                  break;
                case 1: //killmod
                  try {
                    writeRemoteRaw(server.killModule(new Integer(data)));
                  } catch (Exception e) {
                    writeRemote("Invalid number format!");
                  }
                  break;
                case 2: //killallmod
                  server.killAllModules();
                  break;
                case 3: //showmod
                  writeRemoteRaw(server.listModules());
                  break;
                case 4: //showclient
                  writeRemoteRaw(server.listClients());
                  break;
                case 5: //killclient
                  try {
                    writeRemoteRaw(server.killClient(new Integer(data)));
                  } catch (Exception e) {
                    writeRemote("Invalid number format!");
                  }
                  break;
                case 6: //quit
                  System.out.println("Admin client disconnected: " + new java.util.Date().toString());
                  writeRemote("Exiting admin client...");
                  writeRemoteRaw("<command>close</command>");
                  remoteAdminClient.close();
                  break;
                case 7: //help
                  printCommands();
                  break;
                case 8: //killserver
                  terminate = true;
                  writeRemote("Server is going down...");
                  System.out.println("Admin client disconnected: " + new java.util.Date().toString());
                  writeRemote("Exiting admin client...");
                  writeRemoteRaw("<command>close</command>");
                  remoteAdminClient.close();
                  break;
              }
              oxp.reset();
            }
          }
        } while(remoteAdminClient.isConnected());
        
      }
      System.out.println("Server is going down...");
      server.kill();
    } catch (Exception e) {
      System.out.println("Remote admin console thread closing...");
    } finally {
      server.kill();
    }
  }
  /** Wraps a string message in <i>&lt;_msg&gt;</i> tags
   *  to send to the connected console; this calls
   *  writeRemoteRaw() which actually writes to the
   *  to the stream. 2006-08-11 Kyle Reed
   */
  private void writeRemote(String s) {
    writeRemoteRaw("<_msg>" + s + "</_msg>");
  }
  /** Sends a string to the connected admin console.
   *  2006-08-11 Kyle Reed
   */
  private void writeRemoteRaw(String s) {
    try {
      remoteAdminClient.write(ByteBuffer.wrap(s.getBytes()));
    } catch (Exception e) {
      System.out.println("Error writing to remote console. Closing remote connection.");
      try {
        remoteAdminClient.close();
      } catch (Exception e2) {
        //Nothing
      }
    }
  }
  /** Sends a list of the admin commands to the
   *  connected admin console.  2006-08-11 Kyle Reed
   */
  private void printCommands() {
    writeRemote("Command ............................Description");
    writeRemote("-------                             -----------");
    writeRemote("<loadmod>module_name</loadmod> .....Load a module");
    writeRemote("<killmod>module_id</killmod> .......Kill a running module by Module ID");
    writeRemote("<killallmod></killallmod> ..........Kill all modules");
    writeRemote("<showmod></showmod> ................Display the running modules");
    writeRemote("<showclient></showclient> ..........Display connected clients");
    writeRemote("<killclient>client_id</killclient> .Kill a connected client by Client ID");
    writeRemote("<quit></quit> ......................Quit the admin console");
    writeRemote("<help></help> ......................Print command list");
    writeRemote("<killserver></killserver> ..........Terminate server and quit");
  }
  
  /** Returns a reference to the server contained
   *  by this AdminConsole. 2006-08-11 Kyle Reed
   */
  public Server getServer() {
    return server;
  }
  /** Returns a reference to the Server's SocketChannel.
   *  2006-08-11 Kyle Reed
   */
  public ServerSocketChannel getServerSocket() {
    return remoteAdminServer;
  }
}