package Server;

import java.nio.ByteBuffer;
import java.nio.channels.SelectionKey;
import java.nio.channels.SocketChannel;
import java.util.Date;
/** Abstract encapsulation of an incoming socket connection.
 *  Maintains the state of the connection within the server.
 *  Provides the API for all incomming connections.
 *  Can be inherited to create specialized connection types.
 *  2006/08/10 Kyle Reed
 */
public abstract class ConnectionContainer {
  //Shared for all containers
  private static int totalConnections = 0;
  static Server server;
  public static void setServer(Server s) {server = s;}
  
  protected SelectionKey myKey;
  protected int idNum;
  protected String idName;
  protected Date startTime;
  protected int challengeNum;
  protected boolean connected;
  protected SocketChannel io;
  protected ByteBuffer bb;
  protected byte[] ba;
  
  //CONSTRUCTOR
  /** Creates a generic container for an incoming connection.
   *  This class is abstract and cannot be directly instantiated.
   *  Sets the object UID, the default parameters. 2006/08/10 Kyle Reed
   */
  public ConnectionContainer(SelectionKey key) {
    idNum = ++totalConnections;
    idName = "<undef>";
    myKey = key;
    challengeNum = (int)(Math.random() * 10000);
    startTime = new Date();
    io = (SocketChannel)myKey.channel();
  }
  
  //REFLEXIVE METHODS
  /** Tells the container to read its input from <i>io</i> socket and process
   *  it accordingly. This was added to simplify routing in the server
   *  and allowing 200 lines of routing functions to be replaced with
   *  about 20 lines in the container. 2006/08/10 Kyle Reed
   */
  public abstract void handleInput();
  /** Writes a string out on the SocketChannel output stream that belongs to the
   *  SelectionKey passed to the constructor.  Basically it writes <i>to</i> the
   *  connection. 2006/08/10 Kyle Reed
   */
  public void writeToSelf(String s) {
    try {
      io.write(ByteBuffer.wrap(s.getBytes()));
    } catch (Exception e) {
      System.out.println("Failed write to connection: " + idNum);
      selfDestruct();
    }
  }
  /** Send the <i>close</i> tag to the output stream to let the client know
   *  that the connection is about to be closed. 2006/08/10 Kyle Reed
   */
  public void sendTerminator() {
    writeToSelf("<command>close</command>");
  }
  /** Abstract method to get the object's data structures ready for
   *  garbage collection. 2006/08/10 Kyle Reed
   */
  public abstract void selfDestruct();
  
  //SETS AND GETS
  /** Returns the container's UID. 2006/08/10 Kyle Reed
   */
  public int getIdNum() {
    return idNum;
  }
  /** Returns the object's challenge number. 2006/08/10 Kyle Reed
   */
  public int getChallengeNum() {
    return challengeNum;
  }
  /** Returns the object's connected time in seconds. 2006/08/10 Kyle Reed
   */
  public long getUptime() {
    Date temp = new Date();
    long uptime = temp.getTime() - startTime.getTime();
    uptime /= 1000;
    return uptime;
  }
  /** Returns the object's display name. 2006/08/10 Kyle Reed
   */
  public String getName() {
    return idName;
  }
  /** Sets the object's display name. 2006/08/10 Kyle Reed
   */
  public void setName(String s) {
    idName = s;
  }
}
