package Server;
import java.nio.ByteBuffer;
import java.nio.channels.SelectionKey;
import java.util.ArrayList;
import java.util.HashMap;
import simsa.xml.XMLParser;

/** Contains all of the state information for a SIMSA module
 *  as well as handles inbound messages according to module
 *  specifications with a module parser. 2006/08/10 Kyle Reed
 */
public class ModuleContainer extends ConnectionContainer {
  //private Module mod;
  private HashMap<Integer, ClientContainer> registeredClients;
  private XMLParser oxp, ixp;
  private String myType;
  
  private ArrayList<String> vo_c = new ArrayList<String>();
  private ArrayList<String> vo_n = new ArrayList<String>();
  private ArrayList<String> vi_c = new ArrayList<String>();
  private ArrayList<String> vi_n = new ArrayList<String>();
  
  //CONSTRUCTORS
  /** Creates a module container for an incoming connection.
   *  See Connection Container documentation.
   * 2006/08/10 Kyle Reed
   */
  public ModuleContainer(SelectionKey key) {
    super(key);
    myType = "";
    connected = false;
    
    //valid messages when connected
    vo_c.add("_pkt"); //0
    vi_c.add("to");  //0
    vi_c.add("_msg"); //1
    
    //valid messages when not connected
    vo_n.add("connect");    //0
    vo_n.add("disconnect"); //1
    vi_n.add("name");  //0
    vi_n.add("mtype"); //1
    vi_n.add("mid");   //2
    vi_n.add("key");   //3
    
    oxp = new XMLParser(vo_n);
    ixp = new XMLParser(vi_n);
    registeredClients = new HashMap<Integer, ClientContainer>();
  }
  
  //CLIENT FUNCTIONS
  /** Adds a clientContainer reference to registration list.
   *  2006/08/10 Kyle Reed
   */
  public void registerClient(ClientContainer cc) {
    if(!registeredClients.containsKey(cc.getIdNum()))
      registeredClients.put(cc.getIdNum(), cc);
  }
  /** Removes a client from the module's registeredClients
   *  list. 2006/08/10 Kyle Reed
   */
  public void removeClient(Integer key) {
    if(registeredClients.containsKey(key))
      registeredClients.remove(key);
  }
  
  /** Implements the message handling section of a module.
   *  This functions processes incoming data accoring to the
   *  SIMSA spec for modules. 2006/08/10 Kyle Reed
   */
  public void handleInput() {
    bb = ByteBuffer.allocate(server.MAX_BUFFER_SIZE);
    bb.clear();
    int readNum;
    try {
      readNum = io.read(bb);
      bb.flip();
      ba = bb.array();
      
      //The read function returns -1 on EOF (Stream closed)
      if(readNum == -1) throw new Exception("Connection Closed.");
      
      for(int i = 0; i < readNum; i++) {
        oxp.parse((char)ba[i]);
        if(oxp.hasData()) {
          String data = oxp.getData();
          if(connected) {
            switch (oxp.getTagType()) {
              case 0: //_pkt
              {
                String recips = "";
                String msg = "";
                for(char inner_temp : data.toCharArray()) {
                  ixp.parse(inner_temp);
                  if(ixp.hasData()) {
                    switch(ixp.getTagType()) {
                      case 0: //to
                        recips = ixp.getData();
                        break;
                      case 1: //_msg
                        msg = ixp.getData();
                        break;
                    }
                    ixp.reset();
                  }
                }
                ixp.reset(); //prepare inner parser for new data
                
                if(!recips.equals("") && !msg.equals("")) {
                  if(recips.charAt(0) == '0') { //handle broadcast ID
                    for(ClientContainer cc : registeredClients.values()) {
                      if(cc != null)
                        cc.writeToSelf("<_msg>" + msg + "</_msg>");
                    }
                  } else if(recips.charAt(0) == '!') { //handle NOT broadcast ID
                    int ignoreNum = Integer.parseInt(recips.substring(1));
                    for(ClientContainer cc : registeredClients.values()) {
                      if(cc != null && cc.getIdNum() != ignoreNum)
                        cc.writeToSelf("<_msg>" + msg + "</_msg>");
                    }
                  } else {
                    ClientContainer cc;
                    for(String client : recips.split(",")) {
                      cc = registeredClients.get(Integer.parseInt(client));
                      if(cc != null)
                        cc.writeToSelf("<_msg>" + msg + "</_msg>");
                    }
                  }
                }
              }
              break;
            }
          } else {
            switch (oxp.getTagType()) {
              case 0: //connect
              {
                String tempKey = "";
                for(char inner_temp : data.toCharArray()) {
                  ixp.parse(inner_temp);
                  if(ixp.hasData()) {
                    switch(ixp.getTagType()) {
                      case 0: //name
                        idName = ixp.getData();
                        break;
                      case 1: //mtype
                        myType = ixp.getData();
                        break;
                      case 3: //key
                        tempKey = ixp.getData();
                    }
                    ixp.reset();
                  }
                }
                ixp.reset();
                
                if(!idName.equals("<undef>") && !myType.equals("")) {
                  //validate unique names
                  //or name and key
                  
                  //initialize new parsers
                  oxp = new XMLParser(vo_c);
                  ixp = new XMLParser(vi_c);
                  connected = true;
                  writeToSelf("<conack><mid>" + idNum + "</mid>" +
                      "<key>" + challengeNum + "</key></conack>");
                } else {
                  //fix the error message
                  writeToSelf("<error><name>oops</name></error>");
                }
              }
              break;
              case 1: //disconnect
              {
                int mid = 0;
                int chNum = 0;
                for(char inner_temp : data.toCharArray()) {
                  ixp.parse(inner_temp);
                  if(ixp.hasData()) {
                    switch(ixp.getTagType()) {
                      case 2: //mid
                        mid = ixp.getDataAsInt();
                        break;
                      case 3: //key
                        chNum = ixp.getDataAsInt();
                        break;
                    }
                    ixp.reset();
                  }
                }
                ixp.reset();
                
                if(mid != 0 && chNum != 0) {
                  server.killModuleAuth(mid, chNum);
                } else {
                  //error tag here
                }
              }
              server.killModule(idNum);
              break;
            }
          }
          oxp.reset();
        }
      }
    } catch (Exception e) {
      System.out.println("Connection to module: " + idNum + " has failed.  Disconnecting...");
      server.killModule(idNum);
    }
  }
  
  //REFLEXIVE FUNCTIONS
  /**This function attempts to cleanly bring down a module and its clients,
   * set the calling opject to null after this function has been called to
   * ensure that the GC picks it up.
   * 2006/08/10 Kyle Reed
   */
  public void selfDestruct() {
    try {
      //kill all connected clients
      Object[] ccArr = registeredClients.values().toArray();
      for(int i = 0; i < ccArr.length; i++) {
        ClientContainer cc;
        cc = (ClientContainer)ccArr[i];
        cc.writeToSelf("<_msg>The module you were connected to has been disconnected</_msg>");
        server.killClient(cc.getIdNum());
      }
      
      myKey.channel().close();
      myKey.cancel();
      
      System.out.println("Module #" + idNum + " has been disconnected.");
    } catch(Exception e) {
      //The connection was already closed if this block is reached
    }
  }
  
  //SETS AND GETS
  /** Returns the mtype of the module as was passed to in
   *  by the <i>connect</i>. 2006/08/10 Kyle Reed
   */
  public String getType() {
    return myType;
  }
  /** Returns the number of clients registered with the module.
   *  2006/08/10 Kyle Reed
   */
  public int getNumberOfClients() {
    return registeredClients.size();
  }
}
