package Server;
import java.net.*;
import java.nio.ByteBuffer;
import java.nio.channels.*;
import java.util.*;
import simsa.module.Module;

/** Reference implementation of the SIMSA Server that contain
 *  the abstracted connections for clients and modules.  The
 *  main server loop handles client and module registration and
 *  message routing.  2006-08-15 Kyle Reed
 */
public class Server {
  //SERVER VARIABLES
  private Selector socketListener;
  private ServerSocketChannel clientPort, modulePort;
  private SelectionKey clientPortKey, modulePortKey, incomingKey;
  private SocketChannel incomingChannel;
  private Set<SelectionKey> selectedKeys;
  private Iterator<SelectionKey> selectionKeyIter;
  private Thread server;
  private boolean alive;
  
  //CONSTANTS
  public final int MAX_BUFFER_SIZE = 1024;
  
  //REGISTRATION VARIABLES
  HashMap<Integer, ClientContainer> clientList;
  HashMap<Integer, ModuleContainer> moduleList;
  
  //CONTRUCTORS
  /** Initializes registration lists and starts the main
   *  server thread. 2006-08-15 Kyle Reed
   */
  public Server() {
    alive = true;
    clientList = new HashMap<Integer, ClientContainer>();
    moduleList = new HashMap<Integer, ModuleContainer>();
    if(open()) {
      ConnectionContainer.setServer(this);
      run();
    }
  }
  
  //MAIN SERVER FUNCTIONS
  /** Returns true if all of the server sockets have been
   *  successfully bound to the Selector. 2006-08-15 Kyle Reed
   */
  private boolean open() {
    //returns true if all sockets are successfully bound
    try {
      socketListener = Selector.open();
      
      clientPort = ServerSocketChannel.open();
      clientPort.socket().bind(new InetSocketAddress(12345));
      clientPort.configureBlocking(false);
      clientPortKey = clientPort.register(socketListener, SelectionKey.OP_ACCEPT);
      
      modulePort = ServerSocketChannel.open();
      modulePort.socket().bind(new InetSocketAddress(12346));
      modulePort.configureBlocking(false);
      modulePortKey = modulePort.register(socketListener, SelectionKey.OP_ACCEPT);
      return true;
    } catch (Exception e) {
      System.out.println("Exception was thrown creating server connections.");
      return false;
    }
  }
  /** The main server loop that listens for incoming
   *  messages and processes them. 2006-08-15 Kyle Reed
   */
  private void run() {
    server = new Thread(new Runnable() {
      public void run() {
        try {
          System.out.println("Starting server thread...");
          while(isAlive()) {
            getActiveKeys();
            while(hasIncoming()) {
              processIncoming();
            }
          }
        } catch (ClosedSelectorException cse) {
          System.out.println("Server listener has closed.");
        } catch (Exception e) {
          System.out.println("Server thread encountered an error:");
          System.out.println(e.getMessage());
          e.printStackTrace();
        } finally {
          kill();
        }
      }
    });
    server.start();
  }
  /** Get the set of keys that have incoming data and
   *  puts them into a set for processing.
   *  2006/08/10 Kyle Reed
   */
  private void getActiveKeys() throws Exception {
    socketListener.select();
    selectedKeys = socketListener.selectedKeys();
    selectionKeyIter = selectedKeys.iterator();
  }
  /** Returns true if there is another key in the Selection
   *  set that needs to be processed. 2006-08-15 Kyle Reed
   */
  private boolean hasIncoming() {
    if(selectionKeyIter.hasNext()) {
      incomingKey = selectionKeyIter.next();
      selectionKeyIter.remove();
      return true;
    }
    return false;
  }
  /** Processes an incoming message from the selector, it will
   *  register new clients or call the <i>handleInput()</i> method
   *  of the appropriate ConnectionContainer. 2006-08-15 Kyle Reed
   */
  private void processIncoming() {
    if(incomingKey == clientPortKey) {
      registerClient(); //REGISTER NEW CLIENT
    } else if(incomingKey == modulePortKey) {
      registerModule(); //REGISTER NEW MODULE
    } else {
      if(!incomingKey.isValid()) { //REMOVE BAD CONNECTIONS
        System.out.println("Invalid connection! Removing connection:" + incomingKey.toString());
        removeIncoming();
        return;
      } else {
        ((ConnectionContainer)incomingKey.attachment()).handleInput();
      }
    }
  }
  /** Removes a client/module from the server by calling
   *  the <i>selfDestruct()</i> which will cancel its
   *  registration with the Selector and remove itself
   *  from the registration lists.
   */
  private void removeIncoming() {
    ConnectionContainer cc = (ConnectionContainer)incomingKey.attachment();
    //This will work until we combine the lists
    clientList.remove(cc.getIdNum());
    moduleList.remove(cc.getIdNum());
    cc.selfDestruct();
    cc = null;
  }
  
  //REGISTRATION
  /** Adds a module to the module list and assigns it
   *  to a new ModuleContainer. 2006-08-15 Kyle Reed
   */
  private void registerModule() {
    try {
      if(incomingKey.isAcceptable()) {
        incomingChannel = modulePort.accept();
        incomingChannel.configureBlocking(false);
        incomingKey = incomingChannel.register(socketListener, SelectionKey.OP_READ);
      }
      
      ModuleContainer mc = new ModuleContainer(incomingKey);
      incomingKey.attach(mc); //bind the new continer to the selector key
      moduleList.put(mc.getIdNum(), mc);
      System.out.println("Registered new module: " + mc.getIdNum());
    } catch (Exception e) {
      System.out.println("An error occured when registering a module.");
    }
  }
  /** Adds a client to the client list and assigns it
   *  to a new ClientContainer. 2006-08-15 Kyle Reed
   */
  private void registerClient() {
    try {
      if(incomingKey.isAcceptable()) {
        incomingChannel = clientPort.accept();
        incomingChannel.configureBlocking(false);
        incomingKey = incomingChannel.register(socketListener, SelectionKey.OP_READ);
      }
      ClientContainer cc = new ClientContainer(incomingKey);
      incomingKey.attach(cc); //bind the new continer to the selector key
      clientList.put(cc.getIdNum(), cc);
      System.out.println("Registered new client: " + cc.getIdNum());
    } catch (Exception e) {
      System.out.println("An error occured when registering a client.");
    }
  }
  
  //EXTERNAL SERVER API FUNCTIONS
  public Selector getSelector() {
    return socketListener;
  }
  public synchronized boolean isAlive() {
    //This get called from the WatchDog so sync access
    return alive;
  }
  /**
   * Kills the server by setting the
   * alive boolean value to false,
   * breaking out of main while loop
   */
  public synchronized void kill() {
    alive = false;
  }
  
  //FUNCTIONS BELOW ARE MAINLY FOR THE ADMINCONSOLE
  //MODULE SPECIFIC FUNCTIONS
  /** Loads a module (by class name) dynamically to be run in the same VM
   *  as the server.  The module must be in the class path
   *  for the server so that it can be found. 2006-08-15 Kyle Reed
   */
  public String loadModule(String mod) {
    try {
      Class c = Class.forName(mod);
      Module m = (Module)c.newInstance();
      m.connect("localhost", 12346);
      Thread modRun = new Thread(m); //actually starts the module
      modRun.start();
      return "<_msg>Module loaded: " + mod + "</_msg>";
    } catch (InstantiationException ie) {
      System.out.println("Cannot create instance of this module.");
    } catch (ClassNotFoundException cnfe) {
      System.out.println("No class can be found by that name.");
    } catch (Exception e) {
      System.out.println("An error occured loading module.");
    }
    return "<_msg>Module " + mod + " couldn't be loaded.</_msg>";
  }
  /** Returns a formatted string of the loaded modules. 2006-08-15 Kyle Reed
   */
  public String listModules() {
    String temp = "";
    if(!moduleList.isEmpty()) {
      temp = String.format("<_msg>%-11s %-14s %-13s %-7s %-9s</_msg>", "Module ID",
          "Uptime (sec)", "Module Name",  "MType", "# Clients");
      temp += String.format("<_msg>%-11s %-14s %-13s %-7s %-9s</_msg>",  "---------",
          "------------",  "-----------", "-----", "---------");
      for(ModuleContainer mc : moduleList.values()) {
        temp += String.format("<_msg>%-11s %-14s %-13s %-7s %-9s</_msg>", mc.getIdNum(),
            mc.getUptime(), mc.getName(), mc.getType(), mc.getNumberOfClients());
      }
    } else {
      temp = "<_msg>There are no modules loaded</_msg>";
    }
    return temp;
  }
  /** Calls the <i>selfDestruct()</i> method of the
   *  ModuleContainer to remove the Module with the
   *  given ID from thse server. 2006-08-15 Kyle Reed
   */
  public String killModule(int id) {
    String temp = "<_msg>Module not found.</_msg>";
    if(moduleList.containsKey(id)) {
      ModuleContainer mc = moduleList.get(id);
      moduleList.remove(id);
      mc.sendTerminator();
      mc.selfDestruct();
      mc = null;
      temp = "<_msg>Module Unloaded</_msg>";
    }
    return temp;
  }
  /** Calls the <i>selfDestruct()</i> method of the
   *  ModuleContainer to remove the Module with the
   *  given ID from thse server, this allows modules to
   *  pass their Challange key to disconnect themselves.
   *  2006-08-15 Kyle Reed
   */
  public String killModuleAuth(int id, int chNum) {
    String temp = "<_msg>Module not found.</_msg>";
    if(moduleList.containsKey(id)) {
      ModuleContainer mc = moduleList.get(id);
      if(mc.getChallengeNum() == chNum) {
        moduleList.remove(id);
        mc.sendTerminator();
        mc.selfDestruct();
        mc = null;
        temp = "<_msg>Module Unloaded</_msg>";
      }
    }
    return temp;
  }
  /** Kills all of the modules on the server.
   *  2006-08-15 Kyle Reed
   */
  public void killAllModules() {
    //have to use array not for each otherwise we get
    //a modified iterator exception
    Object[] mcArr = moduleList.values().toArray();
    for(int i = 0; i < mcArr.length; i++) {
      ModuleContainer mc;
      mc = (ModuleContainer)mcArr[i];
      mc.sendTerminator();
      mc.selfDestruct();
      mc = null;
    }
    moduleList.clear();
  }
  
  //CLIENT SPECIFIC FUNCTIONS
  /** Returns a formatted string of the loaded client. 2006-08-15 Kyle Reed
   */
  public String listClients() {
    String temp = "";
    if(!clientList.isEmpty()) {
      temp = String.format("<_msg>%-11s %-14s %-13s %-6s</_msg>", "Client ID",
          "Uptime (sec)", "Client Name", "On Mod");
      temp += String.format("<_msg>%-11s %-14s %-13s %-6s</_msg>",  "---------",
          "------------",  "-----------", "------");
      for(ClientContainer cc : clientList.values()) {
        try {
          if(cc.connectedModule() != null) {
            temp += String.format("<_msg>%-11s %-14s %-13s %-6s</_msg>", cc.getIdNum(),
                cc.getUptime(), cc.getName(), cc.connectedModule().getIdNum());
          } else {
            temp += String.format("<_msg>%-11s %-14s %-13s %-6s</_msg>", cc.getIdNum(),
                cc.getUptime(), cc.getName(), "NONE");
          }
        } catch (Exception e) {
          System.out.println("Error printing client list..." + e.getMessage());
        }
      }
    } else {
      temp = "<_msg>There are no clients connected</_msg>";
    }
    return temp;
  }
  /** Calls the <i>selfDestruct()</i> method of the
   *  ClientContainer to remove the Client with the
   *  given ID from the server. 2006-08-15 Kyle Reed
   */
  public String killClient(int id) {
    String temp = "<_msg>Client not found.</_msg>";
    if(clientList.containsKey(id)) {
      ClientContainer cc = clientList.get(id);
      clientList.remove(id);
      cc.sendTerminator();
      cc.selfDestruct();
      cc = null;
      temp = "<_msg>Client Disconnected</_msg>";
    }
    return temp;
  }
  /** Calls the <i>selfDestruct()</i> method of the
   *  ClientContainer to remove the Client with the
   *  given ID from thse server, this allows modules to
   *  pass their Challange key to disconnect themselves.
   *  2006-08-15 Kyle Reed
   */
  public String killClientAuth(int id, int chNum) {
    String temp = "<_msg>Client not found.</_msg>";
    if(clientList.containsKey(id)) {
      ClientContainer cc = clientList.get(id);
      if(cc.getChallengeNum() == chNum) {
        clientList.remove(id);
        cc.sendTerminator();
        cc.selfDestruct();
        cc = null;
        temp = "<_msg>Client Disconnected</_msg>";
      }
    }
    return temp;
  }
}